#include <BLR.h>
#include <data.h>
#include <simpleBLR.h>
#include <cstdlib>
#include <dirent.h>
#include <cstring>
#include <map>
#include <cassert>
#include <set>

void BLR::train ( Data * data , double lambda , const char * dir , int bias , double eps ) {

  char cmd[512];
  sprintf(cmd,"mkdir %s",dir);
  system(cmd);

  std::map<int,double> lambdas,ncnt;
  int P = 0;

  DLOOP ( doc , (*data) ) {
    FLOOP ( f , (*doc) ) P = std::max( P , (*f.id)+1 );
    LLOOP ( l , (*doc) ) lambdas[*l]++;
    LLOOP ( l , (*doc) ) ncnt[*l]++;
  }

  double* x2 = new double[P], sumx2 = 0;
  memset( x2 , 0 , P*sizeof(double));

  DLOOP ( doc , (*data) ) {
    FLOOP ( f , (*doc) ) x2[*f.id] += (*f.val)*(*f.val);
  }

  for ( int i = 0; i < P; ++i ) 
    if ( x2[i] > 0 ) sumx2 += 1.0/x2[i];


  for ( std::map<int,double>::iterator it = lambdas.begin();
	it != lambdas.end(); it++ ) {
    double cnt = it->second;
    if ( lambda > 0 ) {
      it->second = lambda;
    }
    else {
      double pr = cnt * 1.0 / data->ndocs;
      it->second = pr*(1-pr)*P/sumx2 * -lambda;
    }
  }

  double F = 0;
  #pragma omp parallel for reduction(+: F)
  for ( int i = 0; i < data->llen; ++i ) if ( data->labels[i] == 254) {
    char fname[512];
    int lab = data->labels[i];
    sprintf(fname,"%s/%d.gz",dir,lab);

    simpleBLR S ( data , lambdas[lab], bias );
    double F1 = S.solve ( lab , eps );
    S.writeBinary(fname);
    F += F1;
  }
  printf("\n OBJECTIVE = %lf\n",F);

  delete[] x2;
}


void BLR::classify ( Data *data , const char *dir , const char *output ) {

  struct dirent **namelist;
  int k = scandir(dir, &namelist, 0, alphasort) , n = data->ndocs;

  double * bestsc = new double[n], *sc = new double[n];
  int * pr = new int[n];
  for ( int i = 0; i < n; ++i ) 
    bestsc[i] = -2e9 , pr[i] = -1;

  printf("Classifying..\n");
  const int MAXL = 1000000;
  int *tp = new int[MAXL], *fp = new int[MAXL], *fn = new int[MAXL];

  memset( tp , 0 , sizeof(int)*MAXL );
  memset( fp , 0 , sizeof(int)*MAXL );
  memset( fn , 0 , sizeof(int)*MAXL );
  
  for ( int i = 0; i < k; ++i ) {
    if ( strstr( namelist[i]->d_name , "gz" ) ) {
      char fname[512];
      sprintf(fname,"%s/%s",dir,namelist[i]->d_name);
      simpleBLR S( fname );
      assert( S.label > 0 );
      #pragma omp parallel for
      for ( int d = 0; d < n; ++d ) {
	sc[d] = S.classify( data->docs+d );
	if ( sc[d] > bestsc[d] ) {
	  bestsc[d] = sc[d];
	  pr[d] = S.label;
	}
	if ( sc[d] > 0 ) {
	  if ( data->docs[d].labels[0] == S.label ) tp[S.label]++;
	  else fp[S.label]++;
	}
	else
	  if ( data->docs[d].labels[0] == S.label ) fn[S.label]++;
      } 
    }
    free(namelist[i]);
  }

  FILE * f = fopen(output,"w");
  for ( int i = 0; i < n; ++i ) 
    fprintf ( f, "%d %d\n" , pr[i] , data->docs[i].labels[0] );
  fclose ( f );

  /*
  for ( int i = 0; i < MAXL; ++i ) if ( tp[i] + fp[i] + fn[i] > 0 ) {
    double prec = (tp[i]+fp[i]>0?tp[i]*1.0/(tp[i] + fp[i]):0);
    double rec = (tp[i]+fn[i]>0?tp[i]*1.0/(tp[i]+fn[i]):0);
    double f1 = 0;
    if ( prec > 1e-9 && rec > 1e-9 ) f1 = 2*prec*rec/(prec+rec);
    printf("%d -> %lf\n", i , f1 );
  }

  */
  delete[] bestsc;
  delete[] sc;
  delete[] pr;
  delete[] tp;
  delete[] fp;
  delete[] fn;

  if ( k > 0 ) free(namelist);

  /*

  int n = data->ndocs, cut = 20;

  std::set< std::pair<double,int> > *bestsc = new std::set< std::pair<double,int> >[n];

  printf("Classifying..\n");

  for ( int i = 0; i < data->llen; ++i ) {
    int L = data->labels[i];
    char fname[512];
    sprintf(fname,"%s/%d.gz",dir,L);
    if ( fexist(fname) == 0 ) continue;

    simpleBLR S( fname );
    //    #pragma omp parallel for
    for ( int d = 0; d < n; ++d ) {
      double sc = S.classify( data->docs+d );
      bestsc[d].insert( std::make_pair(-sc,S.label) );
      if ( bestsc[d].size() > cut ) bestsc[d].erase( --bestsc[d].end() );
    } 
  }

  FILE * f = fopen(output,"w");
  for ( int i = 0; i < n; ++i ) {
    for ( int j = 0; j < data->docs[i].llen; ++j ) 
      fprintf ( f , "%s%d", j ? "," : "" , data->docs[i].labels[j] );
    for ( std::set< std::pair<double,int> >::iterator it = bestsc[i].begin(); it != 
	    bestsc[i].end(); ++it ) {
      fprintf ( f, " %lf:%d" , -it->first , it->second );
    }
    fprintf ( f , "\n");
  }
  fclose ( f );

  delete[] bestsc;
  */  
}
