#include <DCD.h>
#include <data.h>
#include <simpleDCD.h>
#include <cstdlib>
#include <dirent.h>
#include <cstring>
#include <map>
#include <cassert>
#include <set>

void DCD::train ( Data * data , double lambda , const char * dir , int bias , double eps ) {

  char cmd[512];
  sprintf(cmd,"mkdir %s",dir);
  system(cmd);

  std::map<int,double> lambdas;
  double x2 = 0;
  int P ;

  DLOOP ( doc , (*data) ) {
    FLOOP ( f , (*doc) ) x2 += (*f.val) * (*f.val), P = std::max( P , (*f.id) );
    LLOOP ( l , (*doc) ) lambdas[*l]++;
  }

  for ( std::map<int,double>::iterator it = lambdas.begin();
	it != lambdas.end(); it++ ) {
    double c = it->second;
    if ( lambda > 0 )
      it->second = lambda;
    else
      it->second = c/data->ndocs * ( 1 - c/data->ndocs ) * x2 / P* -lambda ;
  }

  double F = 0;
  #pragma omp parallel for reduction(+: F)
  for ( int i = 0; i < data->llen; ++i ) {
    printf("{%d}", data->labels[i] );
    char fname[512];
    int lab = data->labels[i];
    sprintf(fname,"%s/%d.gz",dir,lab);

    simpleDCD S ( data , lambdas[lab], bias );
    //    if ( fexist(fname) )
    //      S.readBinary(fname);
    F += S.solve ( lab , eps );
    S.writeBinary(fname);
  }
  printf("\n OBJECTIVE = %lf\n",F);
}


void DCD::classify ( Data *data , const char *dir , const char *output ) {

  struct dirent **namelist;
  int k = scandir(dir, &namelist, 0, alphasort) , n = data->ndocs;

  double * bestsc = new double[n], *sc = new double[n];
  int * pr = new int[n];
  for ( int i = 0; i < n; ++i ) 
    bestsc[i] = -2e9 , pr[i] = -1;

  printf("Classifying..\n");

  for ( int i = 0; i < k; ++i ) {
    if ( strstr( namelist[i]->d_name , "gz" ) ) {
      char fname[512];
      sprintf(fname,"%s/%s",dir,namelist[i]->d_name);
      //      printf("%s,",fname);
      
      simpleDCD S( fname );
      #pragma omp parallel for
      for ( int d = 0; d < n; ++d ) {
	sc[d] = S.classify( data->docs+d );
	if ( sc[d] > bestsc[d] ) {
	  bestsc[d] = sc[d];
	  pr[d] = S.label;
	}
      } 
    }
    free(namelist[i]);
  }
  
  FILE * f = fopen(output,"w");
  assert(f);
  for ( int i = 0; i < n; ++i ) 
    fprintf ( f, "%d %d\n" , pr[i] , data->docs[i].labels[0] );

  fclose ( f );

  delete[] bestsc;
  delete[] sc;
  delete[] pr;

  if ( k > 0 ) free(namelist);

  /*

  int n = data->ndocs, cut = 20;

  std::set< std::pair<double,int> > *bestsc = new std::set< std::pair<double,int> >[n];

  printf("Classifying..\n");

  for ( int i = 0; i < data->llen; ++i ) {
    int L = data->labels[i];
    char fname[512];
    sprintf(fname,"%s/%d.gz",dir,L);
    if ( fexist(fname) == 0 ) continue;

    simpleDCD S( fname );
    //    #pragma omp parallel for
    for ( int d = 0; d < n; ++d ) {
      double sc = S.classify( data->docs+d );
      bestsc[d].insert( std::make_pair(-sc,S.label) );
      if ( bestsc[d].size() > cut ) bestsc[d].erase( --bestsc[d].end() );
    } 
  }

  FILE * f = fopen(output,"w");
  for ( int i = 0; i < n; ++i ) {
    for ( int j = 0; j < data->docs[i].llen; ++j ) 
      fprintf ( f , "%s%d", j ? "," : "" , data->docs[i].labels[j] );
    for ( std::set< std::pair<double,int> >::iterator it = bestsc[i].begin(); it != 
	    bestsc[i].end(); ++it ) {
      fprintf ( f, " %lf:%d" , -it->first , it->second );
    }
    fprintf ( f , "\n");
  }
  fclose ( f );

  delete[] bestsc;
  */

}
