#include <GBLR.h>
#include <data.h>
#include <simpleGBLR.h>
#include <simpleGDCD.h>
#include <cstdlib>
#include <dirent.h>
#include <cstring>
#include <vector>
#include <cmath>
#include <algorithm>
#include <set>

extern void printAcc( const char * );

void GBLR::train ( Data * data , double lambda , const char * dir , HInfo * hinfo , double eps , Data * testdata , const char *resfile , char *method , int cut ) {
  char cmd[512];
  sprintf(cmd,"mkdir -p %s",dir);
  system(cmd);

  std::map<int,int> labels;
  std::vector<int> all = hinfo->Labels();
  for ( int i = 0 ; i < all.size(); ++i ) labels[all[i]] = 0;
  for ( int i = 0; i < data->llen; ++i ) labels[ data->labels[i] ] = 1;

  double F = 1e9, prevF = 0;

  for ( int j = 1; j <= 100; ++j ) {

    prevF = F, F = 0;
    #pragma omp parallel for reduction(+: F)
    for ( int i = 0; i < all.size(); ++i ) if ( labels[all[i]]%2 == j%2 ) {
      int L = all[i];
	char fname[512];
	sprintf(fname,"%s/%d.gz",dir,L);
	if ( strcmp(method,"blr") == 0 ) {
	  int bias = 1;
	  simpleGBLR S ( data , lambda , hinfo , bias );
	  double F1 = S.solve ( dir , L , eps );
	  F += F1;
	  S.writeBinary(fname);
	}
	else if ( strcmp(method,"dcd") == 0 ) {
	  int bias = 1;
	  simpleGDCD S ( data , lambda , hinfo , bias );
	  double F1 = S.solve ( dir , L , eps );
	  F += F1;
	  S.writeBinary(fname);
	}
    }
    printf(" OBJECTIVE = %lf\n",F);

    char output[1024];
    sprintf( output , "%s.%d", resfile , j );
    this->classify( testdata , hinfo , dir , output ,method , cut );

    sprintf( output , "%s.%d.train", resfile , j );
    this->classify( data , hinfo , dir , output ,method , cut );
    //printAcc( output );
  }
}



void GBLR::classify ( Data *data , HInfo * hinfo , const char *dir , const char *output , char *method , int cut ) {
  int n = data->ndocs;

  std::set< std::pair<double,int> > *bestsc = new std::set< std::pair<double,int> >[n];

  printf("Classifying..\n");

  for ( int i = 0; i < data->llen; ++i ) {
    int L = data->labels[i];
    char fname[512];
    sprintf(fname,"%s/%d.gz",dir,L);
    //printf("%s\n", fname );
    if( fexist(fname) == 0 ) continue;

    if ( strcmp(method,"blr") == 0 ) {
      simpleGBLR S( L , fname , hinfo );
      #pragma omp parallel for
      for ( int d = 0; d < n; ++d ) {
	double sc = S.classify( data->docs+d );
	bestsc[d].insert( std::make_pair(-sc,S.label) );
	if ( bestsc[d].size() > cut ) bestsc[d].erase( --bestsc[d].end() );
      } 
    }
    else if ( strcmp(method,"dcd") == 0 ) {
      simpleGDCD S( L , fname , hinfo );
      #pragma omp parallel for
      for ( int d = 0; d < n; ++d ) {
	double sc = S.classify( data->docs+d );
	bestsc[d].insert( std::make_pair(-sc,S.label) );
	if ( bestsc[d].size() > cut ) bestsc[d].erase( --bestsc[d].end() );
      } 
    }
  }

  FILE * f = fopen(output,"w");
  for ( int i = 0; i < n; ++i ) {
    for ( int j = 0; j < data->docs[i].llen; ++j ) 
      fprintf ( f , "%s%d", j ? "," : "" , data->docs[i].labels[j] );
    for ( std::set< std::pair<double,int> >::iterator it = bestsc[i].begin(); it != 
	    bestsc[i].end(); ++it ) {
      fprintf ( f, " %lf:%d" , -it->first , it->second );
    }
    fprintf ( f , "\n");
  }
  fclose ( f );

  delete[] bestsc;
  
}
