#include <HBLR.h>
#include <data.h>
#include <simpleHBLR.h>
#include <simpleHDCD.h>
#include <cstdlib>
#include <dirent.h>
#include <cstring>
#include <vector>
#include <cmath>
#include <algorithm>

extern void printAcc( const char * );

void HBLR::train ( Data * data , double lambda , const char * dir , HInfo * hinfo , double eps , Data * testdata , const char *resfile , char *method ) {
  char cmd[512];
  sprintf(cmd,"mkdir -p %s",dir);
  system(cmd);

  std::map<int,int> levels;
  std::vector<int> labels = hinfo->Labels();
  for ( int i = 0 ; i < labels.size(); ++i ) levels[labels[i]] = hinfo->Level(labels[i]);


  double F = 1e9, prevF = 0;

  for ( int j = 0; j < 100; ++j ) {

    prevF = F, F = 0;
    for ( int ii = 0; ii < 10; ++ii ) {

    #pragma omp parallel for reduction(+: F)
    for ( int i = 0; i < labels.size(); ++i ) if ( levels[labels[i]] == ii ) {
	char fname[512];
	sprintf(fname,"%s/%d.gz",dir,labels[i]);

	if ( strcmp(method,"blr") == 0 ) {
	  int bias = hinfo->isLeaf(labels[i]);
	  simpleHBLR S ( data , lambda , hinfo , bias );
	  double F1 = S.solve ( dir , labels[i] , eps );
	  F += F1;
	  S.writeBinary(fname);
	}
	else if ( strcmp(method,"dcd") == 0 ) {
	  int bias = 1;
	  printf("{%d}",labels[i]);
	  fflush(stdout);
	  simpleHDCD S ( data , lambda , hinfo , bias );
	  double F1 = S.solve ( dir , labels[i] , eps );
	  F += F1;
	  S.writeBinary(fname);
	}
      }
    }

    printf(" OBJECTIVE = %lf\n",F);

    char output[1024];
    sprintf( output , "%s.%d", resfile , j );
    this->classify( testdata , hinfo , dir , output ,method );

    //    sprintf( output , "%s.%d.train", resfile , j );
    //    this->classify( data , hinfo , dir , output ,method );
    //printAcc( output );
  }
}


void HBLR::classify ( Data *data , HInfo * hinfo , const char *dir , const char *output , char *method ) {
  int n = data->ndocs;

  double * bestsc = new double[n], *sc = new double[n];
  int * pr = new int[n];
  for ( int i = 0; i < n; ++i ) 
    bestsc[i] = -2e9 , pr[i] = -1;

  printf("Classifying..\n");
  std::vector<int> labels = hinfo->Labels();

  for ( int i = 0; i < labels.size(); ++ i ) if ( hinfo->isLeaf(labels[i]) ) {
    char fname[512];
    sprintf(fname,"%s/%d.gz",dir,labels[i]);

    if ( strcmp(method,"blr") == 0 ) {
      simpleHBLR S( labels[i] , fname , hinfo );
      #pragma omp parallel for
      for ( int d = 0; d < n; ++d ) {
	sc[d] = S.classify( data->docs+d );
	if ( sc[d] > bestsc[d] ) {
	  bestsc[d] = sc[d];
	  pr[d] = S.label;
	}
      } 
    }
    else if ( strcmp(method,"dcd") == 0 ) {
      simpleHDCD S( labels[i] , fname , hinfo );
      #pragma omp parallel for
      for ( int d = 0; d < n; ++d ) {
	sc[d] = S.classify( data->docs+d );
	if ( sc[d] > bestsc[d] ) {
	  bestsc[d] = sc[d];
	  pr[d] = S.label;
	}
      } 
    }
  }

  FILE * f = fopen(output,"w");
  int tot = 0, correct = 0;
  for ( int i = 0; i < n; ++i ) {
    tot++;
    correct += (pr[i] == data->docs[i].labels[0]);
    fprintf ( f, "%d %d\n" , pr[i] , data->docs[i].labels[0] );
  }
  fclose ( f );
  printf("\n Accuracy = %lf\n", correct*1.0/tot);

  printAcc( output );

  delete[] bestsc;
  delete[] sc;
  delete[] pr;
  
}
