#include <HBLRAk.h>
#include <data.h>
#include <simpleHBLRAk.h>
#include <cstdlib>
#include <dirent.h>
#include <cstring>
#include <vector>
#include <cmath>
#include <algorithm>

extern void printAcc( const char * );

void HBLRAk::setup ( Data * data  , const char * dir , HInfo * hinfo , double lambda ) {
  
  std::vector<int> labels = hinfo->Labels();
  std::map<int,double> cnt, ap , bp;

  int P = 0;

  DLOOP ( doc , (*data) ) {
    FLOOP ( f , (*doc) ) P = std::max( P , (*f.id)+1 );
    LLOOP ( l , (*doc) ) cnt[*l]++;
  }
  
  double* x2 = new double[P], sumx2 = 0;
  memset( x2 , 0 , P*sizeof(double));

  DLOOP ( doc , (*data) ) {
    FLOOP ( f , (*doc) ) x2[*f.id] += (*f.val)*(*f.val);
  }

  for ( int i = 0; i < P; ++i ) 
    if ( x2[i] > 0 ) sumx2 += 1.0/x2[i];

  for ( int i = 0 ; i < int(labels.size()); ++i ) {
    if( hinfo->isLeaf(labels[i]) == false ) continue;
    double pr = cnt[labels[i]]*1.0 / data->ndocs;
    double _ap = P;
    double _bp = 1/(pr*(1-pr)) * sumx2;
    std::vector<int> p = hinfo->getPath(labels[i]);
    for ( int j = 0; j < int(p.size()); ++j ) 
      ap[p[j]] += _ap, bp[p[j]] += _bp;
  }
  
  #pragma omp parallel for
  for ( int i = 0; i < int(labels.size()); ++i ) {
    int L = labels[i];
    nodeParam n;
    n.label = L;
    n.P = P;
    n.w = (double*) realloc(n.w,sizeof(double)*P);
    memset(n.w,0,sizeof(double)*P);
    n.aprior = ap[L];
    n.bprior = bp[L] * lambda;
    n.sigma = 1;
    n.alpha = n.aprior / n.bprior;
    //printf("%lf\n", n.alpha );
    char fname[128];
    sprintf(fname,"%s/%d.gz",dir,L);
    n.writeBinary(fname);
  }

  printf("root alpha = %lf\n", ap[0]/bp[0]);
  delete[] x2;
}



void HBLRAk::train ( Data * data  , const char * dir , HInfo * hinfo , double eps , Data * testdata , const char *resfile ) {
  char cmd[512];
  sprintf(cmd,"mkdir -p %s",dir);
  system(cmd);

  std::vector<int> labels = hinfo->Labels();
  double alphas[10000] = {};

  for ( int j = 0; j < 100; ++j ) {
    double F = 0;

    #pragma omp parallel for reduction(+: F)
    for ( int i = 0; i < int(labels.size()); ++i ) {
      //if ( !hinfo->isLeaf(labels[i]) ) continue;
      char fname[512];
      sprintf(fname,"%s/%d.gz",dir,labels[i]);
      simpleHBLRAk S ( data , hinfo );
      F += S.solve ( dir , labels[i] , eps );
      S.writeBinary(fname);
      alphas[ labels[i] ] = S.param->alpha;
    }
    //    for ( int i = 0; i < 1000; ++i ) if ( alphas[i] > 0 ) printf(" %d -> %lf\n", i , alphas[i] );

    printf(" OBJECTIVE = %lf\n",F);

    char output[1024];
    sprintf( output , "%s.%d", resfile , j );
    this->classify( testdata , hinfo , dir , output );
    printAcc( resfile );
  }

}


void HBLRAk::classify ( Data *data , HInfo * hinfo , const char *dir , const char *output ) {
  int n = data->ndocs;

  double * bestsc = new double[n], *sc = new double[n];
  int * pr = new int[n];
  for ( int i = 0; i < n; ++i ) 
    bestsc[i] = -2e9 , pr[i] = -1;

  printf("Classifying..\n");
  std::vector<int> labels = hinfo->Labels();

  for ( int i = 0; i < int(labels.size()); ++ i ) if ( hinfo->isLeaf(labels[i]) ) {
    char fname[512];
    sprintf(fname,"%s/%d.gz",dir,labels[i]);
    
    simpleHBLRAk S( labels[i] , fname , hinfo );
    #pragma omp parallel for
    for ( int d = 0; d < n; ++d ) {
      sc[d] = S.classify( data->docs+d );
      if ( sc[d] > bestsc[d] ) {
	bestsc[d] = sc[d];
	pr[d] = S.param->label;
      }
    } 
  }

  FILE * f = fopen(output,"w");
  int tot = 0, correct = 0;
  for ( int i = 0; i < n; ++i ) {
    tot++;
    correct += (pr[i] == data->docs[i].labels[0]);
    fprintf ( f, "%d %d\n" , pr[i] , data->docs[i].labels[0] );
  }
  fclose ( f );
  printf("\n Accuracy = %lf\n", correct*1.0/tot);

  delete[] bestsc;
  delete[] sc;
  delete[] pr;
  
}
