#include <data.h>
#include <cstdlib>
#include <cstring>
#include <zlib.h>
#include <set>
#include <map>
#include <vector>
#include <algorithm>
#include <sys/stat.h>
#include <cstdlib>
#include <ctime>

int fexist(const char *filename) {
  struct stat s;
  if (stat (filename, &s))  return 0;
  return S_ISREG(s.st_mode);
}


Doc::Doc() {
  fid = labels = NULL;
  fval = NULL;
  flen = llen = 0;
}

void Doc::parse ( Doc * d ) {
  docid = d->docid;
  flen = d->flen;
  llen = d->llen;
  fid = new int[flen];
  fval = new float[flen];
  labels = new int[llen];
  memcpy( fid , d->fid , sizeof(int)*flen);
  memcpy( fval , d->fval , sizeof(float)*flen);
  memcpy( labels , d->labels , sizeof(int)*llen);
}


void Doc::parse ( char * line ) {
  char *str, *token;
  flen = 0, llen = 0;
  for( str = line; *str; str++ ) 
    flen += (*str == ' ' ),
      llen += ( *str == ',');

  // -2 for #, docid, +1 for comma
  flen -= 2, llen += 1;
  fid = new int[flen];
  fval = new float[flen];
  labels = new int[llen];
  flen = llen = 0;

  str = line;
  while ( *str && *str != ' ' ) {
    if ( *str == ',' ) str++;
    sscanf( str, "%d" , &labels[llen++] );
    while( str && *str != ' ' && *str != ',' ) str++;
  }

  while ( (*str) == ' ' && (*str) != '#' ) str++;
  token = strtok ( str ," ");
  while (token != NULL) {
    if ( token[0] == '#' ) {
      token = strtok( NULL , " ");
      sscanf( token , "%d" , &docid );
      break;
    }
    sscanf( token ,  "%d:%f", &fid[flen] , &fval[flen] );
    flen++;
    token = strtok( NULL , " ");
  }  
}

void Doc::writeBinary ( gzFile f ) {
  gzwrite( f , &llen , sizeof(int) );
  gzwrite( f , &flen , sizeof(int) );
  gzwrite( f , &docid , sizeof(int) );
  gzwrite( f , labels , llen*sizeof(int) );
  gzwrite( f , fid , flen*sizeof(int) );
  gzwrite( f , fval , flen*sizeof(float) );
}

void Doc::readBinary ( gzFile f ) {
  gzread( f , &llen , sizeof(int) );
  gzread( f , &flen , sizeof(int) );
  gzread( f , &docid , sizeof(int) );
  fid = new int[flen];
  fval = new float[flen];
  labels = new int[llen];
  gzread( f , labels , llen*sizeof(int) );
  gzread( f , fid , flen*sizeof(int) );
  gzread( f , fval , flen*sizeof(float) );
}

void Doc::print () {
  for ( int i = 0; i < llen; ++i )
    printf("%s%d",i?",":"",labels[i]);
  for ( int i = 0; i < flen; ++i )
    printf(" %d:%g", fid[i] , fval[i] );
  printf(" # %d\n", docid );
}

Doc::~Doc(){
  if ( fid ) delete [] fid;
  if ( fval ) delete [] fval;
  if ( labels ) delete [] labels;
}





void Data::parse ( Data *data , int sel ) {
  llen = data->llen;
  labels = new int[llen];
  memcpy(labels,data->labels,sizeof(int)*llen);
  std::map<int,int> cnts;
  std::vector<int> d;

  std::vector<int> v;
  for ( int i = 0; i < data->ndocs; ++i )
    v.push_back(i);
  
  srand ( 0 );//time(NULL) );
  //  std::random_shuffle( v.begin() , v.end() );

  for ( int j = 0; j < data->ndocs; ++j ) {
    int i = v[j];
    int L = data->docs[i].labels[0];
    if ( cnts[L] < sel ) {
      cnts[L]++;
      d.push_back(i);
    }
  }

  ndocs = d.size();
  docs = new Doc[ndocs];
  for ( int i = 0; i < ndocs; ++i ) 
    docs[i].parse( &data->docs[d[i]] );
}

void Data::parse ( const char *filename ) {
  int c , lines = 0;
  static char line[50<<20];
  FILE * fptr = fopen(filename,"r");

  if ( fptr == NULL ) {
    perror("Cannot open data file");
    exit(0);
  }

  while ( (c=fgetc(fptr)) != -1 ) 
    lines += (c=='\n');
  ndocs = lines;
  docs = new Doc[ndocs];

  rewind(fptr);
  lines = 0;
  while ( true ) {
    line[0] = 0;
    fgets ( line , (50<<20)-1 , fptr );
    if ( line[0] == 0 && feof(fptr) ) break;
    docs[lines++].parse( line );
  }
  
  std::set<int> unique_labels;
  for ( int i = 0; i < ndocs; ++i ) 
    for ( int j = 0; j < docs[i].llen; ++j )
      unique_labels.insert(docs[i].labels[j]);
  llen = unique_labels.size();

  labels = (int*) malloc(sizeof(int)*llen);
  llen = 0;
  for ( std::set<int>::iterator it = unique_labels.begin();
	it != unique_labels.end(); ++it )
    labels[llen++] = (*it);
}

Data::Data() {
  docs = NULL;
  labels = NULL;
  ndocs = 0;
}

void Data::print () {
  for ( int i = 0; i < ndocs; ++i ) 
    docs[i].print();
}

void Data::writeBinary ( const char * filename ) {
  gzFile f = gzopen( filename , "wb" );
  if ( f == NULL ) {
    perror("Cannot open data file for writing");
    exit(0);
  }

  gzwrite( f , &ndocs , sizeof(int) );
  gzwrite( f , &llen , sizeof(int) );
  gzwrite( f , labels , sizeof(int)*llen );
  for ( int i = 0; i < ndocs; ++i )
    docs[i].writeBinary( f );
  gzclose( f );
}

void Data::readBinary ( const char * filename ) {
  gzFile f = gzopen( filename , "rb" );
  if ( f == NULL ) {
    perror("Cannot open data file for reading");
    exit(0);
  }

  gzread( f , &ndocs , sizeof(int) );
  gzread( f , &llen , sizeof(int) );
  docs = new Doc[ndocs];
  labels = new int[llen];
  gzread( f , labels , sizeof(int)*llen );
  for ( int i = 0; i < ndocs; ++i )
    docs[i].readBinary( f );
  gzclose( f );
}

Data::~Data() {
  if ( docs ) delete [] docs;
  if ( labels ) delete[] labels;
}
