#include <lbfgssolver.h>
#include <lbfgsinterface.h>
#include <cstdio>
#include <ctime>

extern double __eps;
extern int __iterations;
extern int __iprint;

double evaluate ( void * instance , const double *x , double *g , const int n , const double step ) {
  LBFGSinterface * inst = ( LBFGSinterface * ) instance;
  return inst->fg(x,g,n);
}



int progress ( void *instance , const double * x , const double *g , const double fx , const double xnorm , 
	      const double gnorm , const double step , int n , int k , int ls ) {
  printf("%4d  %4d  %12e  %12e  %12e\n", k , ls , fx , gnorm , step );
  return 0;
}



LBFGSsolver::LBFGSsolver ( double * x , int _n , LBFGSinterface* interface , double eps ) {
  double l1 = 0;
  int print = __iprint;
  ret = -1;
  n = _n;
  f = 2e9;
  lbfgs_parameter_init(&param);    
  if ( __eps > 0 )
    param.epsilon = __eps;
  param.max_iterations = __iterations;

  if ( l1 > 0 ) {
    param.linesearch = LBFGS_LINESEARCH_BACKTRACKING;
    param.orthantwise_c = l1;
    param.orthantwise_start = 0;
    param.orthantwise_end = _n-2;
  }

  if ( print ) 
    printf("%4s  %4s  %12s  %12s  %12s\n", "ITER" , "NFN" , "FUNC" , "GNORM" , "STEP" );

  if ( print ) 
    ret = lbfgs(n,x,&f,evaluate,progress,interface,&param);
  else 
    ret = lbfgs(n,x,&f,evaluate,NULL,interface,&param);

  if ( ret != 0 ) {
    if ( ret == LBFGSERR_ROUNDING_ERROR ) 
      ;//      printf(" ERROR = ROUNDING ERROR\n");
    else if ( ret == LBFGSERR_INVALID_LINESEARCH ) {
      printf(" ERROR ENCOUNTERED WHILE MINIMIZATION\n");
      printf(" ERROR = INVALID LINE SEARCH\n");
    }
    else if ( ret ==  LBFGS_ALREADY_MINIMIZED ){
      //printf(" ERROR ENCOUNTERED WHILE MINIMIZATION\n");
      //printf(" ERROR = ALREADY MINIMIZED\n");
    }
    else {
      printf(" ERROR ENCOUNTERED WHILE MINIMIZATION\n");
      printf(" RETURN ERROR CODE = %d\n", ret );
    }
  }
  else if ( print ) 
    printf(" MINIMIZATION COMPLETE\n");
}
