#include <data.h>
#include <cstdlib>
#include <simpleMLR.h>
#include <simpleBLR.h>
#include <simpleTBLR.h>
#include <BLR.h>
#include <TBLR.h>
#include <simpleHBLR.h>
#include <simpleHDCD.h>
#include <HBLR.h>
#include <hinfo.h>
#include <pthread.h>
#include <DCD.h>
#include <cstring>
#include <cassert>
#include <HBLRAk.h>
#include <GBLR.h>

int __iterations = 200;
double __eps = 1e-5;
int __iprint = 0;

const int nf = 10000000, SZ = 128;
pthread_mutex_t *filelocks;

char trainpath[SZ],testpath[SZ],method[SZ],hpath[SZ],resfile[SZ],paramfolder[SZ],gpath[SZ];
double lambda;
void init( int , char *[] );
void temp( Data * , Data * , HInfo * );
void destroy();
void printAcc( const char * );

double dcdeps=.01;

int main( int argc , char *argv[] ){
  if ( argc != 4 ){
    puts("./exe <dataset> <method> <lambda>");
    exit(0);
  }
  /*
  Data temp;
  temp.parse("a.svm");
  temp.writeBinary("a.gz");
  exit(0);
  */

  init( argc , argv );
  Data *train = new Data(), *test = new Data();
  HInfo *hinfo = NULL, *ginfo = NULL;

  if ( fexist(hpath) ) 
    hinfo = new HInfo(hpath);
  if ( fexist(gpath) )
    ginfo = new HInfo(gpath);

  train->readBinary(trainpath);    
  test->readBinary(testpath);

/*
  Data *subset = new Data();
  subset->parse( train , 10 );
  train = subset;
*/

  /*  
  subset = new Data();
  subset->parse( test , 50 );
  test = subset;
  */
  /*  
  std::vector<int> labels = hinfo->Labels();
  std::map<int,int> cnt;

  DLOOP(d,(*train)){
    LLOOP(l,(*d)) cnt[*l]++;
  }

  for ( int i = 0; i < labels.size(); ++i ) {
    if ( hinfo->isLeaf( labels[i] ) == 1 ) {
      char fname[512];
      sprintf(fname,"%s/%d.gz",paramfolder,labels[i]);
      simpleHDCD N ( labels[i] , fname , hinfo );
      double x = 0;
      for ( int j = 0; j < N.P; ++j ) 
	x += N.w[j]*N.w[j];
      printf("%d %lf %d\n", labels[i] , x , cnt[labels[i]] );
    }
  }
  exit(0);
  */

  int bias = 1;
  if ( strcmp(method,"blr") == 0 ) {
    BLR B;
    B.train( train , lambda , paramfolder , bias , 1e-5 );
    B.classify( test , paramfolder , resfile );
  } 
  else if ( strcmp(method,"mlr") == 0 ) {
    simpleMLR S( train , lambda  , bias );
    S.writeBinaryFolder(paramfolder);
    BLR B;
    B.classify( test , paramfolder , resfile );
    char output[1024];
    sprintf( output , "%s.train", resfile );
    B.classify( train , paramfolder , output );
  }
  else if ( strcmp(method,"dcd") == 0 ) {
    DCD D;
    D.train( train , lambda , paramfolder , bias , dcdeps );
    D.classify( test , paramfolder , resfile );
    //char output[1024];
    //sprintf( output , "%s.train", resfile );
    //D.classify( train , paramfolder , output );
  }
  else if ( strcmp(method,"tblr") == 0 ) {
    TBLR BT;
    BT.train( train , lambda , paramfolder , bias , 1e-5 , test , resfile );
    BT.classify( test ,  paramfolder , resfile );
  }
  else if ( strcmp(method,"hblr") == 0 ) {
    HBLR H;
    H.train( train , lambda , paramfolder , hinfo , 1e-6 , test , resfile , "blr" );
  }
  else if ( strcmp(method,"hdcd") == 0 ) {
    HBLR H;
    H.train( train , lambda , paramfolder , hinfo , dcdeps , test , resfile , "dcd" );
    //H.classify( test , hinfo , paramfolder , resfile , "dcd" );
  }
  else if ( strcmp(method,"hblrak") == 0 ) {
    HBLRAk H;
    H.setup ( train , paramfolder , hinfo , lambda );
    H.train ( train , paramfolder , hinfo , 1e-6 , test , resfile );
  }
  else if ( strcmp(method,"gdcd") == 0 ) {
    GBLR G;
    G.train ( train , lambda , paramfolder , ginfo , dcdeps , test , resfile , "dcd" , 20 );
  }
  else if ( strcmp(method,"gblr") == 0 ) {
    GBLR G;
    G.train ( train , lambda , paramfolder , ginfo , 1e-6 , test , resfile , "blr" , 20 );
  }

  printAcc( resfile );
  destroy();
}







void init( int argc , char *argv[] ){
  filelocks = new pthread_mutex_t[nf];
  for ( int i = 0; i < nf; ++i )
    pthread_mutex_init(&filelocks[i],NULL);

  sprintf(trainpath,"datasets/%s/%s.train", argv[1] , argv[1] );
  sprintf(testpath,"datasets/%s/%s.test", argv[1] , argv[1] );
  sprintf(hpath,"datasets/%s/%s.hierarchy", argv[1] , argv[1] );
  sprintf(gpath,"datasets/%s/%s.graph", argv[1] , argv[1] );
  sprintf(method,"%s",argv[2]);
  sscanf(argv[3],"%lf",&lambda);
  sprintf(paramfolder,"params/%s/%s",argv[1],argv[2]);
  sprintf(resfile,"preds/%s/%s/%s.pred",argv[1],argv[2],argv[2]);
  
}


void destroy() {
  for ( int i = 0; i < nf; ++i )
    pthread_mutex_destroy(&filelocks[i]);
}

struct Tab {
  int tp,fp,fn;
  Tab(){tp=fp=fn=0;}
};

void printAcc( const char * resfile ){

  std::map<int,Tab> mp;

  FILE *f = fopen(resfile,"r");
  if ( f ) {
    int tot = 0, correct = 0;
    for ( int a,b; fscanf(f,"%d %d",&a,&b) == 2; ){
      if ( a == b ) mp[a].tp += 1, correct += 1;
      else mp[a].fp += 1, mp[b].fn += 1;
      tot++;
    }
    double f1 = 0;
    for ( std::map<int,Tab>::iterator it = mp.begin();
	  it != mp.end(); ++it ) {
      double p = it->second.tp*1.0/(it->second.tp+it->second.fp);
      double r = it->second.tp*1.0/(it->second.tp+it->second.fn);
      double f = ( p+r>1e-9 ? (2*p*r/(p+r)) : 0 );
      f1 += f;
    }
    printf(" Accuracy = %lf\n Ma-f1 = %lf\n", correct*1.0/tot , f1/mp.size() );
  }
}




void temp( Data * tr , Data *te , HInfo *hinfo ){
  for ( int i = 0; i < tr->ndocs; ++i ) 
    tr->docs[i].labels[0] = hinfo->Parent( tr->docs[i].labels[0] );
  for ( int i = 0; i < te->ndocs; ++i ) 
    te->docs[i].labels[0] = hinfo->Parent( te->docs[i].labels[0] );

  for ( std::map<int, std::vector<int> >::iterator it = hinfo->children.begin(); it != hinfo->children.end(); ++it ) {
    int cnt = 0;
    const std::vector<int>& temp = it->second;
    for ( int i = 0; i < int(temp.size()); ++i) if ( hinfo->isLeaf(temp[i]) == 0 ) cnt++;
    if ( cnt ) {
      printf("%d %d", it->first , cnt );
      for ( int i = 0; i < int(temp.size()); ++i) if ( hinfo->isLeaf(temp[i]) == 0 ) printf(" %d", temp[i]);
      puts("");
    }
  }

  //tr->writeBinary("tr.gz");
  //te->writeBinary("te.gz");
  exit(0);
}



void kNN( const Data& train , const Data& test ) {
  /*
  #pragma omp parallel for
  for ( int i = 0; i < test.ndocs; ++i ) {
    Doc *te = &test.docs[i];
    double bestscore = -(1<<30);
    int pred = 0;
    DLOOP(tr,train){
      int a = 0, b = 0;
      double sc = 0;
      while ( a < tr->flen && b < te->flen ) {
	if ( tr->fid[a] == te->fid[b] ) 
	  sc += tr->fval[a] * te->fval[b], a++, b++;
	else if ( tr->fid[a] < te->fid[b] )
	  a++;
	else
	  b++;
      }
      if ( sc > bestscore ) {
	bestscore = sc;
	pred = tr->labels[0];
      }
    }
    printf("%d %d %d\n", te->docid , te->labels[0] , pred);
  }
  exit(0);
  */
}
