#include <nodeParam.h>
#include <cstdlib>
#include <zlib.h>
#include <cstring>
#include <cmath>
#include <vector>
#include <cstdio>
#include <lbfgsinterface.h>


nodeParam::nodeParam(){
  w = (double*) malloc(sizeof(1));
  w[0] = 0;
  P = 1; 
  aprior = bprior = sigma = alpha = ap = bp = 1;
  label = -1;
}


void nodeParam::writeBinary( const char *fname ) {
  pthread_mutex_lock( &filelocks[label] );

  gzFile f = gzopen( fname, "wb" );
  if ( f == NULL ) {
    perror("Cannot open data file for writing");
    exit(0);
  }


  gzwrite( f , &label , sizeof(int) );
  gzwrite( f , &aprior , sizeof(double) );
  gzwrite( f , &bprior , sizeof(double) );
  gzwrite( f , &sigma , sizeof(double) );
  gzwrite( f , &alpha , sizeof(double) );
  gzwrite( f , &ap , sizeof(double) );
  gzwrite( f , &bp , sizeof(double) );
  gzwrite( f , &P , sizeof(int) );
  gzwrite( f , w , sizeof(double)*P );
  gzclose(f);

  pthread_mutex_unlock( &filelocks[label] );
}


void nodeParam::readBinary( int lockl , const char *fname ) {
  pthread_mutex_lock( &filelocks[lockl] );

  gzFile f = gzopen( fname, "rb" );
  if ( f == NULL ) {
    perror("Cannot open data file for reading");
    exit(0);
  }

  gzread( f , &label , sizeof(int) );
  gzread( f , &aprior , sizeof(double) );
  gzread( f , &bprior , sizeof(double) );
  gzread( f , &sigma , sizeof(double) );
  gzread( f , &alpha , sizeof(double) );
  gzread( f , &ap , sizeof(double) );
  gzread( f , &bp , sizeof(double) );
  gzread( f , &P , sizeof(int) );
  w = (double*) realloc(w,sizeof(double)*P);
  gzread( f , w , sizeof(double)*P );
  gzclose(f);
  pthread_mutex_unlock( &filelocks[lockl] );
}


double nodeParam::norm() {
  double n = 0;
  for ( int i = 0; i < P; ++i )
    n += w[i]*w[i];
  return n;
}


void nodeParam::print( int x) {
  printf(" label = %d\n aprior = %lf\n bprior = %lf\n sigma = %lf\n alpha = %lf\n P = %d ap = %lf bp = %lf\n", 
	 label, aprior, bprior , sigma , alpha , P, ap , bp );
  if ( x ) 
    for ( int i = 0; i < P; ++i ) printf(" %d:%lf", i , w[i]);
}

nodeParam::~nodeParam(){
  if ( w ) free(w);
}
