#include <simpleBLR.h>
#include <data.h>
#include <zlib.h>
#include <lbfgssolver.h>
#include <set>
#include <cmath>
#include <cassert>
#include <cstdlib>
#include <cstring>

simpleBLR::simpleBLR ( const char *fname ) {
  dataptr = NULL, label = -1, w = NULL, bias = 0;
  lambda = 0;
  P = 0;
  readBinary( fname );
}


simpleBLR::simpleBLR ( Data* _dataptr , double _lambda , int _bias ) {
  dataptr = _dataptr;
  lambda = _lambda;
  label = -1;
  bias = _bias;
  P = 0;

  DLOOP ( doc , *dataptr ) {
    FLOOP ( f , *doc ) P = std::max(P,*f.id+1);
  }

  w = (double*) malloc(sizeof(double)*P);
  memset( w , 0 , sizeof(double)*P );

}


double simpleBLR::solve ( int _label , double eps ) {
  assert( dataptr );
  label = _label;

  LBFGSsolver* solver = new LBFGSsolver( w , P , this , std::min(1e-5,lambda*10) );
  double F = solver->objective();
  delete solver;
  solver = NULL;

  return F;
}


#define INF (1e30)
#define EPS (1e-50)
#define ADD (1e30)
/*
double simpleBLR::fg ( const double *x , double *g , int n ) {
  double F = 0;
  g[0] = 0;
  for ( int i = 1; i < n; ++i ) 
    g[i] = lambda * x[i],
      F += lambda/2 * x[i]*x[i];

  DLOOP ( doc , *dataptr ) {
    double sc = bias ? x[0] : 0;
    FLOOP ( f , *doc ) sc += x[*f.id] * (*f.val);
    sc = isinf(sc=exp(sc)) ? INF: sc;

    double pr = sc / ( 1 + sc );
    int y = ( label == doc->labels[0] );
    FLOOP ( f , *doc ) g[*f.id] += -(y-pr) * (*f.val);
    if ( bias ) g[0] += -(y-pr);
    
    double add = 0;
    if ( y == 1 ) {
      if ( pr < EPS ) add = ADD;
      else add = -log(pr);
    }
    else {
      if ( pr > 1- EPS ) add = ADD;
      else add = -log(1-pr);
    }
    F += add;
  }
  return F;
}
*/

double simpleBLR::fg ( const double *x , double * g , int n ) {
  double F = 0;
  g[0] = 0;
  for ( int i = 1; i < n; ++i ) 
    g[i] = lambda * x[i],
      F += lambda/2 *x[i]*x[i];

  DLOOP ( doc , *dataptr ) {
    int y = -1;
    LLOOP ( l , *doc ) if ( *l == label ) y = +1;
    double wx = bias ? x[0] : 0;
    FLOOP ( f , *doc ) wx += x[*f.id] * (*f.val);
    double sc = exp(y*wx);
    if ( isinf(sc) ) sc = INF;

    double pr = 1 / ( 1 + sc );
    FLOOP ( f , *doc ) g[*f.id] += -y * pr * (*f.val);
    if ( bias ) g[0] += -y*pr;
    
    double add = exp(-y*wx);
    if ( isinf(add) )
      add = ADD;
    else
      add = log(1+add);
    F += add;
  }
  return F;
}


double simpleBLR::classify ( const Doc * doc ) {
  double sc = bias ? w[0] : 0;
  FLOOP ( f , *doc ) if ( *f.id < P ) sc += w[*f.id] * (*f.val); 
  //  sc = isinf(sc=exp(sc)) ? INF: sc;  
  //return sc/(1+sc);
  return sc;
}


void simpleBLR::writeBinary ( const char *fname ) {
  gzFile f = gzopen( fname, "wb" );
  if ( f == NULL ) {
    perror("Cannot open data file for writing");
    exit(0);
  }

  int K = 1;
  gzwrite( f , &K , sizeof(int) );
  gzwrite( f , &P , sizeof(int) );
  gzwrite( f , &bias , sizeof(int) );
  gzwrite( f , &label  , sizeof(int) );
  gzwrite( f , w , sizeof(double)*P );
  gzclose(f);
}


void simpleBLR::readBinary ( const char *fname ) {
  gzFile f = gzopen( fname, "rb" );
  if ( f == NULL ) {
    perror("Cannot open data file for reading");
    exit(0);
  }

  int K = 1;
  gzread( f , &K , sizeof(int) );
  gzread( f , &P , sizeof(int) );
  gzread( f , &bias , sizeof(int) );
  gzread( f , &label , sizeof(int) );
  w = (double*) realloc(w,sizeof(double)*P);
  gzread( f , w , sizeof(double)*P );
  gzclose(f);
  //  printf("{ %d %d %d %d %lf}\n", K , P , bias , label , w[P-1] );
}


simpleBLR::~simpleBLR (){
  free(w);
}

